//=============================================================================
//
//  CLASS Sphere - IMPLEMENTATION
//
//=============================================================================
#define RAYTRACER_SPHERE_CPP

//== INCLUDES =================================================================
#include "Sphere.h"

//== IMPLEMENTATION ==========================================================
//-----------------------------------------------------------------------------
bool Sphere::intersect(
    const Ray       &clRay, 
    double        t0,     
    double        t1,     
    TracingContext  &tContext )  const
{

    /*		
		Strahl-Kugel Schnitt
		Die relevanten (bei Schnitt zu setzenden) Felder im TracingContext sind:

		t          - Strahlparameter des Schnittpunkts
		v3Normal   - Normale im Schnittpunkt
		v3HitPoint - Koordinate des Schnittpunktes
		pclShader  - Verweis auf das Material der Kugel

	*/

	VectorType3 v3Diff = clRay.getOrigin()-m_v3Center;
	
	//The coefficients of the quadratic equation
	double A = clRay.getDir() | clRay.getDir();
	double B = 2*(clRay.getDir() | v3Diff);
	double C = (v3Diff | v3Diff) - m_rRadius*m_rRadius;

	double discriminant = B*B-4*A*C;
	if (discriminant > FLT_EPSILON)
	{
		//check the two solutions
		double sqrtdis = sqrt(discriminant);
		double t = (-B - sqrtdis)/(2*A);
		if (t > t0 && t < t1)
		{
			tContext.m_t = t;
			tContext.m_v3HitPoint = clRay.evaluate(t);
			tContext.m_v3Normal = (tContext.m_v3HitPoint-m_v3Center)*(1/m_rRadius);
			tContext.m_pclShader = getShader();
			return true;
		}
		t = (-B + sqrtdis)/(2*A);
		if (t > t0 && t < t1)
		{
			tContext.m_t = t;
			tContext.m_v3HitPoint = clRay.evaluate(t);
			tContext.m_v3Normal = (tContext.m_v3HitPoint-m_v3Center)*(1/m_rRadius);
			tContext.m_pclShader = getShader();
			return true;
		}
	}
	return false;
}

VectorType2 Sphere::cartesian2Polar(
    VectorType3 &v )  const
{
    /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
    double    z = ( v[2] > 1 ) ? 1 : ( v[2] < -1 ) ? -1 : v[2];
    double    theta = ( double ) acos( z );
    double    phi = ( double ) atan2( v[1], v[0] );
    /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

    if( phi < 0 )
        phi = phi + 2 * ( double ) M_PI;
    return VectorType2( theta, phi );
}


void Sphere::renderGL() const
{

    //Generate s sphere
    int d = 20;
    glPushMatrix();
    glTranslatef( m_v3Center[0], m_v3Center[1], m_v3Center[2] );

    GLfloat step = M_PI / d;
    GLfloat sina, sinb, sinbp, cosbp, cosa, cosb, x, y, z;
    int     count = 0;
    if( m_selected )
    {
        GLfloat mat_ambient[] = { 1, 1, 1 };
        glMaterialfv( GL_FRONT, GL_AMBIENT, mat_ambient );
    }
    else
        m_shader->shadeGL();
    for( GLfloat beta = 0; beta < 2 * M_PI - 0.5 * step; beta += step )
    {
        glBegin( GL_TRIANGLE_STRIP );
        sinb = sin( beta );
        cosb = cos( beta );
        sinbp = sin( beta + step );
        cosbp = cos( beta + step );
        count++;
        glVertex3f( 0, 0, m_rRadius );
        glNormal3f( 0, 0, m_rRadius );
        for( GLfloat alpha = step; alpha < M_PI - 0.5 * step; alpha += step )
        {
            sina = sin( alpha );
            cosa = cos( alpha );
            x = m_rRadius * sina * cosb;
            y = m_rRadius * sina * sinb;
            z = m_rRadius * cosa;
            glVertex3f( x, y, z );
            glNormal3f( x, y, z );
            count++;
            x = m_rRadius * sina * cosbp;
            y = m_rRadius * sina * sinbp;
            z = m_rRadius * cosa;
            glVertex3f( x, y, z );
            glNormal3f( x, y, z );
            count++;
        }

        count++;
        glVertex3f( 0, 0, -m_rRadius );
        glNormal3f( 0, 0, -m_rRadius );
        glEnd();
    }

    glPopMatrix();
}

//translate object
void Sphere::translate(
    float   x,  
    float   y,  
    float   z ) 
{
    m_v3Center[0] = x;
    m_v3Center[1] = y;
    m_v3Center[2] = z;
}

//rotate object
void Sphere::rotate(
    float   /*rotx*/,   
    float   /*roty*/,   
    float   /*rotz*/ )  
{

    //Nothing to do in this case
}
