//=============================================================================
//
//  Class RaytracingArea
//
//=============================================================================
#include <QtGui>
#include "raytracingwidget.h"
#include "xmlParser.h"
#include "Plane.h"
#include "Sphere.h"
#include "Checkerboard.h"
#include "SuperEllipsoid.h"
#include "SuperHyperboloid.h"
#include "SuperToroid.h"
#include "Torus.h"
#include "Metaball.h"
#include "TriangleSoup.h"
#include "ColorT.h"
#include "PointLight.h"
#include "DirectionalLight.h"
#include "Types.h"
#include <time.h>

//Constructor
RaytracingArea::RaytracingArea(
    QWidget *parent )  : QWidget( parent )
{
	m_recursionDepth = 10;
	m_imageSizeX = 25;
	m_imageSizeY = 25;
	m_displayMode = MODE_DISPLAY;

    m_raytracer = new Raytracer();
    m_scene = new SurfaceList();
    m_materials = new std::vector <Shader *>;
    m_materials->clear();
    m_raytracer->getCamera().Reset();
    m_raytracedImage = new QImage( m_imageSizeX, m_imageSizeY,
                                   QImage::Format_RGB32 );
	//Reset Image to white
    for( int i = 0; i < m_raytracedImage->width(); i++ )
    {
        for( int j = 0; j < m_raytracedImage->height(); j++ )
        {
            QColor  color( 255, 255, 255 );
            m_raytracedImage->setPixel( i, j, color.rgb() );
        }
    }
    connect( m_raytracer, SIGNAL( renderImageAndRays( RayArray * ) ),
             this, SLOT( renderImage( RayArray * ) ) );
}


QSize RaytracingArea::minimumSizeHint() const
{
    return QSize( 100, 100 );
}


QSize RaytracingArea::sizeHint() const
{
    return QSize( 400, 400 );
}



/* Refresh Image */
void RaytracingArea::paintEvent(
    QPaintEvent * ) 
{
    /*~~~~~~~~~~~~~~~~~~~~~~~~*/
    QPainter    painter( this );
    /*~~~~~~~~~~~~~~~~~~~~~~~~*/

    if( m_raytracedImage != NULL )
	{
		painter.drawImage( rect(), *m_raytracedImage  );
	}
	
	painter.restore();
}

/* Save Raytraced Image*/
void RaytracingArea::saveImage(
    const QString   &path ) 
{
    QString fileName = path;
    if( fileName.isNull() )
        fileName = QFileDialog::getSaveFileName( this, tr( "Save Image" ), "",
                                                 "Image Files (*.png)" );
    if( !fileName.isEmpty() )
    {
        if( !m_raytracedImage->save( fileName, "png" ) )
        {
            QMessageBox::warning( this, tr( "Save Image" ),
                                  tr( "The image file could not be saved." ),
                                  QMessageBox::Cancel, QMessageBox::NoButton );
            return;
        }
    }
}

/* Loading XML-Scenes */
void RaytracingArea::openScene(
    const QString   &path ) 
{
    m_raytracer->removeAllLights();
    m_scene->removeAllSurfaces();
	m_materials->clear();

    QString fileName = path;
    if( fileName.isNull() )
        fileName = QFileDialog::getOpenFileName( this, tr( "Open Scene" ), "",
                                                 "Scene Files (*.xml)" );
    if( !fileName.isEmpty() )
    {

        //Parse XML-File
        //Open and parse the XML file:
        QByteArray  fn = fileName.toLatin1();
        const char  *fname = fn.data();
        XMLNode     xMainNode = XMLNode::openFileHelper( fname, "scene" );

        //Camera defined?
        int         n = xMainNode.nChildNode( "camera" );
        if( n > 0 )
        {
            float   x, y, z, left, bottom, right, top;
            XMLNode node = xMainNode.getChildNode( "camera" );
            sscanf( node.getAttribute( "pos" ), "%f %f %f", &x, &y, &z );
            m_raytracer->getCamera().setEye( VectorType3( x, y, z ) );
            sscanf( node.getAttribute( "dir" ), "%f %f %f", &x, &y, &z );
            m_raytracer->getCamera().setLookAt( VectorType3( x, y, z ) );
            sscanf( node.getAttribute( "up" ), "%f %f %f", &x, &y, &z );
            m_raytracer->getCamera().setUp( VectorType3( x, y, z ) );
            sscanf( node.getAttribute( "focallength" ), "%f", &x );
            m_raytracer->getCamera().setFocalLength( x );
            sscanf( node.getAttribute( "window" ), "%f %f %f %f", &left,
                    &bottom, &right, &top );
            m_raytracer->getCamera().setWindow( left, bottom, right, top );
        }

        //Materials defined?
        n = xMainNode.nChildNode( "materials" );
        if( n > 0 )
        {
            XMLNode materialNodes = xMainNode.getChildNode( "materials" );
            int     nr = materialNodes.nChildNode( "material" );
            //Search highest ID
			int highestID = nr;
			for( int i = 0; i < nr; i++ )
            {
				XMLNode materialNode = materialNodes.getChildNode( "material", i );
				int tmp;
				sscanf( materialNode.getAttribute( "id" ),
                            "%d", &tmp );
				if ( tmp > highestID )
					highestID = tmp;
			}
			m_materials->resize( highestID );
			
			for( int i = 0; i < nr; i++ )
            {
                XMLNode materialNode = materialNodes.getChildNode( "material", i );
                if( !strcmp( materialNode.getAttribute( "type" ), "phong" ) )
                {
                    float   ar, ag, ab, sr, sg, sb, rr, rg, rb, dr, dg, db;
					int id;
                    float   refractiveIndex, shininess;
                    sscanf( materialNode.getAttribute( "ambient" ), "%f %f %f",
                            &ar, &ag, &ab );
                    sscanf( materialNode.getAttribute( "specular" ), "%f %f %f",
                            &sr, &sg, &sb );
                    sscanf( materialNode.getAttribute( "refractive" ),
                            "%f %f %f", &rr, &rg, &rb );
                    sscanf( materialNode.getAttribute( "refractiveIndex" ),
                            "%f", &refractiveIndex );
                    sscanf( materialNode.getAttribute( "diffuse" ), "%f %f %f",
                            &dr, &dg, &db );
                    sscanf( materialNode.getAttribute( "shininess" ), "%f",
                            &shininess );
                    sscanf( materialNode.getAttribute( "id" ), "%d",
                            &id );
                    PhongShader *
                        phongshader = new PhongShader( ColorType( ar, ag, ab ),
                                                                      ColorType( sr, sg, sb ),
                                                                          ColorType( rr, rg, rb ),
                                                                              refractiveIndex,
                                                                              ColorType( dr, dg, db ),
                                                                                      shininess );
                    (*m_materials)[id] = phongshader;
                }
				int nr = materialNodes.nChildNode( "background" );
				if ( nr > 0 )
				{
					XMLNode backgroundNode = materialNodes.getChildNode( "background" );
					float   r, g, b;
                    sscanf( backgroundNode.getAttribute( "color" ), "%f %f %f",
                            &r, &g, &b );
					m_raytracer->setBackgroundColor( ColorType( r, g, b ) );
				}
				//If no material node is defined set background to black
				else
					m_raytracer->setBackgroundColor( ColorType( 0, 0, 0 ) );
            }
        }

        //Scene defined?
        n = xMainNode.nChildNode( "geometry" );
        if( n > 0 )
        {
            XMLNode geometryNode = xMainNode.getChildNode( "geometry" );
            int     nr = geometryNode.nChildNode( "surface" );
            for( int i = 0; i < nr; i++ )
            {
                XMLNode surfaceNode = geometryNode.getChildNode( "surface", i );
                if( !strcmp( surfaceNode.getAttribute( "type" ), "sphere" ) )
                {
                    float           x, y, z, radius;
                    unsigned int    materialID;
                    XMLNode         sphereNode = surfaceNode.getChildNode( "sphere" );
                    sscanf( sphereNode.getAttribute( "center" ), "%f %f %f", &x,
                            &y, &z );
                    sscanf( sphereNode.getAttribute( "radius" ), "%f", &radius );
                    sscanf( sphereNode.getAttribute( "materialId" ), "%d",
                            &materialID );
                    Sphere * sphere = new Sphere( VectorType3( x, y, z ), radius );
                    if( materialID >= 0 )
                        sphere->setShader( (*m_materials)[materialID] );
                    m_scene->addSurface( sphere );
                }
                else if( !strcmp( surfaceNode.getAttribute( "type" ), "plane" ) )
                {
                    float           x, y, z, nx, ny, nz;
                    unsigned int    materialID;
                    XMLNode         planeNode = surfaceNode.getChildNode( "plane" );
                    sscanf( planeNode.getAttribute( "pos" ), "%f %f %f", &x, &y,
                            &z );
                    sscanf( planeNode.getAttribute( "normal" ), "%f %f %f", &nx,
                            &ny, &nz );
                    sscanf( planeNode.getAttribute( "materialId" ), "%d",
                            &materialID );
                    Plane * plane = new Plane( VectorType3( nx, ny, nz ), VectorType3( x, y, z ) );
                    if( materialID >= 0 )
                        plane->setShader( (*m_materials)[materialID] );
                    m_scene->addSurface( plane );
                }
                else if( !strcmp( surfaceNode.getAttribute( "type" ), "checkerboard" ) )
                {
                    float           x, y, z, x1, y1, z1, x2, y2, z2;
                    unsigned int    materialID1;
					unsigned int    materialID2;
					unsigned int    xRaster, yRaster;
                    XMLNode         planeNode = surfaceNode.getChildNode( "checkerboard" );
                    sscanf( planeNode.getAttribute( "pos" ), "%f %f %f", &x, &y,
                            &z );
					sscanf( planeNode.getAttribute( "pos1" ), "%f %f %f", &x1, &y1,
                            &z1 );
					sscanf( planeNode.getAttribute( "pos2" ), "%f %f %f", &x2, &y2,
                            &z2 );
                    sscanf( planeNode.getAttribute( "materialId1" ), "%d",
                            &materialID1 );
                    sscanf( planeNode.getAttribute( "materialId2" ), "%d",
                            &materialID2 );
                    sscanf( planeNode.getAttribute( "xRaster" ), "%d",
                            &xRaster );
                    sscanf( planeNode.getAttribute( "yRaster" ), "%d",
                            &yRaster );
                    Checkerboard *cb = new Checkerboard( VectorType3( x, y, z ),
						VectorType3( x1, y1, z1 ),
						VectorType3( x2, y2, z2 ), xRaster, yRaster );
                    if( materialID1 >= 0 )
                        cb->setShader( (*m_materials)[materialID1] );
                    if( materialID2 >= 0 )
                        cb->setShader2( (*m_materials)[materialID2] );
                    m_scene->addSurface( cb );
                }
				if( !strcmp( surfaceNode.getAttribute( "type" ), "torus" ) )
                {
                    float           x, y, z, c, a;
                    unsigned int    materialID;
                    XMLNode         sNode = surfaceNode.getChildNode( "torus" );
                    sscanf( sNode.getAttribute( "position" ), "%f %f %f", &x,
                            &y, &z );
                    sscanf( sNode.getAttribute( "c" ), "%f", &c );
                    sscanf( sNode.getAttribute( "a" ), "%f", &a );
                    sscanf( sNode.getAttribute( "materialId" ), "%d",
                            &materialID );
                    Torus *t = new Torus( VectorType3( x, y, z ), c, a );
                    if( materialID >= 0 )
                        t->setShader( (*m_materials)[materialID] );
                    m_scene->addSurface( t );
                }
				if( !strcmp( surfaceNode.getAttribute( "type" ), "superEllipsoid" ) )
                {
                    float           x, y, z, a, b, c, p, q, r;
                    unsigned int    materialID;
                    XMLNode         sNode = surfaceNode.getChildNode( "superEllipsoid" );
                    sscanf( sNode.getAttribute( "position" ), "%f %f %f", &x,
                            &y, &z );
                    sscanf( sNode.getAttribute( "a" ), "%f", &a );
                    sscanf( sNode.getAttribute( "b" ), "%f", &b );
                    sscanf( sNode.getAttribute( "c" ), "%f", &c );
                    sscanf( sNode.getAttribute( "p" ), "%f", &p );
                    sscanf( sNode.getAttribute( "q" ), "%f", &q );
                    sscanf( sNode.getAttribute( "r" ), "%f", &r );
                    sscanf( sNode.getAttribute( "materialId" ), "%d",
                            &materialID );
                    SuperEllipsoid *t = new SuperEllipsoid( VectorType3( x, y, z ), a, b, c, p, q, r );
                    if( materialID >= 0 )
                        t->setShader( (*m_materials)[materialID] );
                    m_scene->addSurface( t );
                }
				if( !strcmp( surfaceNode.getAttribute( "type" ), "superHyperboloid" ) )
                {
                    float           x, y, z, a, b, c, p, q, r;
                    unsigned int    materialID;
                    XMLNode         sNode = surfaceNode.getChildNode( "superHyperboloid" );
                    sscanf( sNode.getAttribute( "position" ), "%f %f %f", &x,
                            &y, &z );
                    sscanf( sNode.getAttribute( "a" ), "%f", &a );
                    sscanf( sNode.getAttribute( "b" ), "%f", &b );
                    sscanf( sNode.getAttribute( "c" ), "%f", &c );
                    sscanf( sNode.getAttribute( "p" ), "%f", &p );
                    sscanf( sNode.getAttribute( "q" ), "%f", &q );
                    sscanf( sNode.getAttribute( "r" ), "%f", &r );
                    sscanf( sNode.getAttribute( "materialId" ), "%d",
                            &materialID );
                    SuperHyperboloid *t = new SuperHyperboloid( VectorType3( x, y, z ), a, b, c, p, q, r );
                    if( materialID >= 0 )
                        t->setShader( (*m_materials)[materialID] );
                    m_scene->addSurface( t );
                }
				if( !strcmp( surfaceNode.getAttribute( "type" ), "superToroid" ) )
                {
                    float           x, y, z, a, b, c, p, q, r, m, n, d, e;
                    unsigned int    materialID;
                    XMLNode         sNode = surfaceNode.getChildNode( "superToroid" );
                    sscanf( sNode.getAttribute( "position" ), "%f %f %f", &x,
                            &y, &z );
                    sscanf( sNode.getAttribute( "a" ), "%f", &a );
                    sscanf( sNode.getAttribute( "b" ), "%f", &b );
                    sscanf( sNode.getAttribute( "c" ), "%f", &c );
                    sscanf( sNode.getAttribute( "p" ), "%f", &p );
                    sscanf( sNode.getAttribute( "q" ), "%f", &q );
                    sscanf( sNode.getAttribute( "r" ), "%f", &r );
                    sscanf( sNode.getAttribute( "m" ), "%f", &m );
                    sscanf( sNode.getAttribute( "n" ), "%f", &n );
                    sscanf( sNode.getAttribute( "d" ), "%f", &d );
                    sscanf( sNode.getAttribute( "e" ), "%f", &e );
                    sscanf( sNode.getAttribute( "materialId" ), "%d",
                            &materialID );
                    SuperToroid *t = new SuperToroid( VectorType3( x, y, z ), a, b, c, p, q, r, m, n, d, e );
                    if( materialID >= 0 )
                        t->setShader( (*m_materials)[materialID] );
                    m_scene->addSurface( t );
                }
				else if( !strcmp( surfaceNode.getAttribute( "type" ), "metaball" ) )
				{
					unsigned int    materialID;
					float			isovalue;
					sscanf( surfaceNode.getAttribute( "materialId" ), "%d", &materialID );
					sscanf( surfaceNode.getAttribute( "isovalue" ), "%f", &isovalue );

					int numBalls = surfaceNode.nChildNode("ball");
					Metaball *metaball = new Metaball(isovalue,numBalls);

					if( materialID >= 0 )
						metaball->setShader( (*m_materials)[materialID] );

					for ( int i = 0; i < numBalls; i++)
					{
						XMLNode         ballNode = surfaceNode.getChildNode( "ball", i );
						float			weight;
						float           x, y, z;
						float			bandwidth;
						sscanf( ballNode.getAttribute( "weight" ), "%f", &weight );
						sscanf( ballNode.getAttribute( "center" ), "%f %f %f", &x, &y, &z );
						sscanf( ballNode.getAttribute( "bandwidth" ), "%f", &bandwidth );
						metaball->addBall( weight, VectorType3(x,y,z), bandwidth );
					}
					m_scene->addSurface( metaball );
				}

				else if( !strcmp( surfaceNode.getAttribute( "type" ), "meshfile" ) )
                {
                    float           x, y, z, sx, sy, sz;
					unsigned int    materialID;
					unsigned int    unitize;
					const char *filename;
                    XMLNode         PlaneRTNode = surfaceNode.getChildNode( "meshfile" );
                    sscanf( PlaneRTNode.getAttribute( "position" ), "%f %f %f", &x, &y, &z );
                    sscanf( PlaneRTNode.getAttribute( "scale" ), "%f %f %f", &sx, &sy, &sz );
                    sscanf( PlaneRTNode.getAttribute( "materialId" ), "%d", &materialID );
                    sscanf( PlaneRTNode.getAttribute( "unitize" ), "%d", &unitize );
                    filename = PlaneRTNode.getAttribute( "filename" );
                    TriangleSoup *ts = new TriangleSoup( filename );
                    if( unitize != 0 )
                        ts->unitizeMax();
                    ts->scale( VectorType3( sx, sy, sz  ) );
                    ts->translate( VectorType3( x, y, z ) );
                    if( materialID >= 0 )
                        ts->setShader( (*m_materials)[materialID] );
                    m_scene->addSurface( ts );
                }
			}
        }

        //Lights defined?
        n = xMainNode.nChildNode( "lights" );
        if( n > 0 )
        {
            XMLNode lightsNode = xMainNode.getChildNode( "lights" );
            int     nr = lightsNode.nChildNode( "light" );
            for( int i = 0; i < nr; i++ )
            {
                XMLNode lightNode = lightsNode.getChildNode( "light", i );
                if( !strcmp( lightNode.getAttribute( "type" ),
                    "directionalLight" ) )
                {
                    float   x, y, z, r, g, b;
                    sscanf( lightNode.getAttribute( "direction" ), "%f %f %f",
                            &x, &y, &z );
                    sscanf( lightNode.getAttribute( "color" ), "%f %f %f", &r,
                            &g, &b );
                    DirectionalLight *
                        dirLight = new DirectionalLight( VectorType3( x, y, z ),
                                                                        ColorType( r, g, b ) );
                    m_raytracer->AddLight( dirLight );
                }
                else if( !strcmp( lightNode.getAttribute( "type" ), "pointLight"
                         ) )
                {
                    float   x, y, z, r, g, b;
                    sscanf( lightNode.getAttribute( "position" ), "%f %f %f",
                            &x, &y, &z );
                    sscanf( lightNode.getAttribute( "color" ), "%f %f %f", &r,
                            &g, &b );
                    PointLight *
                        pointLight = new PointLight( VectorType3( x, y, z ),
                                                                    ColorType( r, g, b ) );
                    m_raytracer->AddLight( pointLight );
                }
            }
        }

        emit    setGLScene( m_raytracer, m_scene );
    }
}

/* Start Rendering */
void RaytracingArea::render()
{
	clock_t start, finish;
	long int duration;

    //Reset Image to white
    for( int i = 0; i < m_raytracedImage->width(); i++ )
    {
        for( int j = 0; j < m_raytracedImage->height(); j++ )
        {
            QColor  color( 255, 255, 255 );
            m_raytracedImage->setPixel( i, j, color.rgb() );
        }
    }

	//Set GL Viewports for Raytracing
	if( m_displayMode == MODE_DISPLAY )
		emit	startRaytracingGL( m_raytracedImage );
	else //start timing
	{
      start = clock ();
	}
    //render the scene
    m_raytracer->render( m_scene, m_recursionDepth, m_raytracedImage, m_displayMode );

	if( m_displayMode == MODE_TIMING )
	{
		finish = clock ();
		duration = finish - start;
		    QMessageBox::about( this, tr( "Timings" ),
				tr( "Time for Raytracing: %1 msec" ).arg(duration) );
	}

    //Reset Rays in OpenGL
    emit    resetGL();
	repaint();
}

/* Refresh Image: Called by Raytracer */
void RaytracingArea::renderImage(
    RayArray *glRays )   
{
	if ( m_displayMode == MODE_DISPLAY )
	{
		emit    renderRaysGL( glRays );
		repaint();
	}
}


void RaytracingArea::updateRecursionDepth( int depth )
{
	m_recursionDepth = depth;
}


void RaytracingArea::updateImageSize( int size )
{
	delete m_raytracedImage;
	m_imageSizeX = size;
	m_imageSizeY = size;
    m_raytracedImage = new QImage( size, size,
                                   QImage::Format_RGB32 );
	//Reset Image to white
    for( int i = 0; i < m_raytracedImage->width(); i++ )
    {
        for( int j = 0; j < m_raytracedImage->height(); j++ )
        {
            QColor  color( 255, 255, 255 );
            m_raytracedImage->setPixel( i, j, color.rgb() );
        }
    }

	repaint();
}

/* Wether or not Ray are rendered */
void RaytracingArea::setDisplayMode(
    int mode )  
{
	if( mode == Qt::Unchecked )
		m_displayMode = MODE_DISPLAY;
	if( mode == Qt::Checked )
		m_displayMode = MODE_TIMING;
}